from dataclasses import dataclass


@dataclass
class Resources:
    # Format to: secret_key
    default_config = """\
# Updates the Flask app config with the variables below.
# If any variable below does not exist in the standard Flask env vars it is created and will be accessible using
# app.config["YOUR_VAR_NAME"]. All variables defined below will be capitalised when imported.

[flask]
APP_NAME = "app"
VERSION = "0.0.1"
SECRET_KEY = "{secret_key}"
DEBUG = false
TESTING = false
SESSION_TIME = 480
ERROR_404_HELP = true
SQLALCHEMY_TRACK_MODIFICATIONS = false
EXPLAIN_TEMPLATE_LOADING = false

# [database.main] is loaded as SQLALCHEMY_DATABASE_URI
# type = mysql / postgresql / sqlite
# if type = sqlite, config parser will ignore username and password
[database]

    [database.main]
    enabled = true
    type = "sqlite"
    database_name = "database"
    location = "db"
    port = ""
    username = "user"
    password = "password"

"""
    # Format to: static_url_path
    default_theme_config = """\
enabled = "yes"
static_folder = "static"
static_url_path = "{static_url_path}"
"""



from functools import wraps

from flask import flash
from flask import redirect
from flask import session
from flask import url_for


class BigAppSecurity:
    """
    A class that provides route function decorators to check for session variables and their values.
    """

    @staticmethod
    def login_required(
            redirect_endpoint: str,
            bool_key_name: str,
            show_message: bool = True,
            message: str = "You need to be logged in to access this page."
    ):

        def login_required_wrapper(func):
            @wraps(func)
            def secure_function(*args, **kwargs):
                if bool_key_name not in session:
                    if show_message:
                        flash(message)
                    return redirect(url_for(redirect_endpoint))

                if not session[bool_key_name]:
                    if show_message:
                        flash(message)
                    return redirect(url_for(redirect_endpoint))

                return func(*args, **kwargs)

            return secure_function

        return login_required_wrapper

    @staticmethod
    def no_login_required(
            redirect_endpoint: str,
            session_bool_key_name: str,
            show_message: bool = True,
            message: str = "You are already logged in."
    ):

        def no_login_required_wrapper(func):
            @wraps(func)
            def secure_function(*args, **kwargs):
                if session_bool_key_name in session:
                    if session[session_bool_key_name]:
                        if show_message:
                            flash(message)
                        return redirect(url_for(redirect_endpoint))

                return func(*args, **kwargs)

            return secure_function

        return no_login_required_wrapper

    @staticmethod
    def permission_required(
            redirect_endpoint: str,
            session_list_key: str,
            permission_needed: str,
            show_message: bool = True,
            message: str = "You don't have the required permissions to access this page."
    ):

        def permission_required(func):
            @wraps(func)
            def secure_function(*args, **kwargs):
                if session_list_key not in session:
                    if show_message:
                        flash(message)
                    return redirect(url_for(redirect_endpoint))

                if permission_needed not in session[session_list_key]:
                    if show_message:
                        flash(message)
                    return redirect(url_for(redirect_endpoint))

                return func(*args, **kwargs)

            return secure_function

        return permission_required



import re
from pathlib import Path

import click

from .resources import Resources


def to_snake_case(string):
    """
    Thank you openai
    """
    # Replace any non-alphanumeric characters with underscores
    string = re.sub(r'[^a-zA-Z0-9]', '_', string)
    # Remove any consecutive underscores
    string = re.sub(r'_{2,}', '_', string)
    # Convert the string to lowercase
    string = string.lower()
    return string


class Sprinkles:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    END = '\033[0m'


@click.group()
def cli():
    pass  # Entry Point


@cli.command("add-bp", help="Create a flask-bigapp blueprint")
@click.option(
    '-f', '--folder',
    nargs=1,
    default="Current Working Directory",
    prompt=(f'\n{Sprinkles.WARNING}(Creation is relative to the current working directory){Sprinkles.END}\n'
            f'Folder to create blueprint in'),
    help='The from_folder to create the blueprint in, defaults to the current working directory'
)
@click.option(
    '-n', '--name',
    nargs=1,
    default="my_new_blueprint",
    prompt='Name of the blueprint to create',
    help='The name of the blueprint to create'
)
@click.option(
    '-urlp', '--url-prefix',
    nargs=1,
    default="/",
    prompt='URL prefix',
    help='The URL Prefix to set for the blueprint'
)
@click.option(
    '-surlp', '--static-url-path',
    nargs=1,
    default="/<blueprint_name>/static",
    prompt='Static URL path',
    help='The URL Prefix to set for the blueprint'
)
def add_blueprint(folder, name, url_prefix, static_url_path):
    cwd = Path.cwd()
    if folder != "Current Working Directory":
        cwd = Path(cwd / folder)
    if not cwd.exists():
        click.echo(
            f"{Sprinkles.FAIL}{folder} does not exist.{Sprinkles.END}")
        return

    name = to_snake_case(name)

    url_prefix = url_prefix if url_prefix != "/" else "/"
    url_prefix = url_prefix if url_prefix.startswith("/") else f"/{url_prefix}"

    if static_url_path == "/<blueprint_name>/static":
        static_url_path = f"/{name}/static"
    else:
        static_url_path = static_url_path if static_url_path.startswith("/") else f"/{static_url_path}"

    bp_folder = cwd / name
    bp_routes_folder = bp_folder / "routes"
    bp_templates_folder = bp_folder / "templates" / name
    bp_static_folder = bp_folder / "static"

    bp_init = bp_folder / "__init__.py"
    bp_config = bp_folder / "config.toml"
    bp_route = bp_routes_folder / "index.py"
    bp_template = bp_templates_folder / "index.html"

    folders = [bp_folder, bp_routes_folder, bp_templates_folder, bp_static_folder]

    for folder in folders:
        if not folder.exists():
            folder.mkdir(parents=True)
        else:
            click.echo(f"{Sprinkles.WARNING}Folder already exists: {folder}, skipping{Sprinkles.END}")

    if not bp_init.exists():
        bp_init.write_text(Resources.blueprint_init)
    else:
        click.echo(f"{Sprinkles.WARNING}__init__ already exists: {bp_init}, skipping{Sprinkles.END}")

    if not bp_config.exists():
        bp_config.write_text(
            Resources.blueprint_config.format(
                name=name,
                url_prefix=url_prefix,
                static_url_path=static_url_path
            )
        )
    else:
        click.echo(f"{Sprinkles.WARNING}Config already exists: {bp_config}, skipping{Sprinkles.END}")

    if not bp_route.exists():
        bp_route.write_text(Resources.blueprint_index_route)
    else:
        click.echo(f"{Sprinkles.WARNING}Route already exists: {bp_route}, skipping{Sprinkles.END}")

    if not bp_template.exists():
        bp_template.write_text(Resources.blueprint_index_template.format(name=name))
    else:
        click.echo(f"{Sprinkles.WARNING}Template already exists: {bp_template}, skipping{Sprinkles.END}")

    click.echo(f"{Sprinkles.OKGREEN}Blueprint created: {bp_folder}{Sprinkles.END}")


@cli.command("rename-bp", help="Rename a flask-bigapp blueprint")
@click.option(
    '-bp', '--blueprint',
    nargs=1,
    prompt=(
            f'\n{Sprinkles.WARNING}(Search of the blueprint is relative to the current working directory){Sprinkles.END}\n'
            f'\nMust be the blueprint from_folder (that contains the __init__.py file)\n'
            f'Blueprint to rename'
    ),
    help='Location of the blueprint to rename'
)
@click.option(
    '-n', '--new-name',
    nargs=1,
    prompt='New name',
    help='The new name to give to the blueprint'
)
def rename_blueprint(blueprint, new_name):
    cwd = Path.cwd()
    blueprint_path = Path(cwd / blueprint)
    if not blueprint_path.exists():
        click.echo(
            f"{Sprinkles.FAIL}The Blueprint [{blueprint}] does not exist in the {blueprint_path.parent} directory.{Sprinkles.END}")
        return

    blueprint_nested_template_path = blueprint_path / "templates" / blueprint_path.name

    new_name = to_snake_case(new_name)

    new_path = blueprint_path.parent / new_name

    if new_path.exists():
        click.echo(f"{Sprinkles.FAIL}The new name you have given already exists.{Sprinkles.END}")
        return

    blueprint_config = blueprint_path / "config.toml"

    with open(blueprint_config, "r") as f:
        config = f.read()

    config = config.replace(blueprint_path.name, new_name)

    with open(blueprint_config, "w") as f:
        f.write(config)

    blueprint_nested_template_path.rename(blueprint_path / "templates" / new_name)
    blueprint_path.rename(blueprint_path.parent / new_name)


@cli.command("add-theme", help="Create a flask-bigapp theme")
@click.option(
    '-f', '--from_folder',
    nargs=1,
    default="Current Working Directory",
    prompt=(f'\n{Sprinkles.WARNING}(Creation is relative to the current working directory){Sprinkles.END}\n'
            f'Folder to create theme in'),
    help='The from_folder to create the theme in, defaults to the current working directory'
)
@click.option(
    '-n', '--name',
    nargs=1,
    default="my_new_theme",
    prompt='Name of the theme to create',
    help='The name of the theme to create'
)
@click.option(
    '-surlp', '--static-url-path',
    nargs=1,
    default="/<theme_name>/static",
    prompt='Static URL path',
    help='The URL Prefix to set for the theme'
)
def add_theme(folder, name, static_url_path):
    import shutil

    cwd = Path.cwd()

    if folder != "Current Working Directory":
        cwd = Path(cwd / folder)
    if not cwd.exists():
        click.echo(
            f"{Sprinkles.FAIL}{folder} does not exist.{Sprinkles.END}")
        return

    name = to_snake_case(name)

    theme_folder = cwd / name

    if theme_folder.exists():
        click.echo(f"{Sprinkles.WARNING}Theme already exists: {theme_folder}, skipping{Sprinkles.END}")
        return

    if static_url_path != "/<theme_name>/static":
        static_url_path = static_url_path if static_url_path.startswith("/") else f"/{static_url_path}"
    else:
        static_url_path = f"/{name}/static"

    temp_them_folder = cwd / "__temp_theme_dir__"
    temp_theme_nested_template_folder = temp_them_folder / "templates" / "__temp_theme_dir__"

    rename_theme_folder = cwd / name
    rename_theme_nested_template_folder = temp_them_folder / "templates" / name

    new_theme_folder = rename_theme_folder
    new_theme_nested_template_folder = new_theme_folder / "templates" / name

    cli_dir = Path(__file__).parent
    cli_theme_dir = cli_dir / "__temp_theme_dir__"

    shutil.copytree(cli_theme_dir, temp_them_folder, dirs_exist_ok=True)

    theme_config = temp_them_folder / "config.toml"
    theme_main = new_theme_nested_template_folder / "extends" / "main.html"

    if not theme_config.exists():
        theme_config.write_text(
            Resources.theme_config.format(
                name=name,
                static_url_path=static_url_path
            )
        )
    else:
        click.echo(f"{Sprinkles.WARNING}Config already exists: {theme_config}, skipping{Sprinkles.END}")

    temp_theme_nested_template_folder.rename(rename_theme_nested_template_folder)
    temp_them_folder.rename(rename_theme_folder)

    with open(theme_main, "r") as f:
        main = f.read()

    main = main.replace("__temp_theme_dir__", name)

    with open(theme_main, "w") as f:
        f.write(main)

    click.echo(f"{Sprinkles.OKGREEN}Theme created: {theme_folder}{Sprinkles.END}")


@cli.command("rename-theme", help="Rename a flask-bigapp theme")
@click.option(
    '-t', '--theme',
    nargs=1,
    prompt=(
            f'\n{Sprinkles.WARNING}(Search of the theme is relative to the current working directory){Sprinkles.END}\n'
            f'Theme to rename'
    ),
    help='Location of the theme to rename'
)
@click.option(
    '-n', '--new-name',
    nargs=1,
    prompt='New name',
    help='The new name to give to the theme'
)
def rename_theme(theme, new_name):
    cwd = Path.cwd()
    theme_path = Path(cwd / theme)
    if not theme_path.exists():
        click.echo(
            f"{Sprinkles.FAIL}The Theme [{theme}] does not exist in the {theme_path.parent} directory.{Sprinkles.END}")
        return

    theme_nested_template_path = theme_path / "templates" / theme_path.name

    new_name = to_snake_case(new_name)
    new_path = theme_path.parent / new_name

    if new_path.exists():
        click.echo(f"{Sprinkles.FAIL}The new name you have given already exists.{Sprinkles.END}")
        return

    theme_config = theme_path / "config.toml"

    with open(theme_config, "r") as f:
        config = f.read()

    config = config.replace(theme_path.name, new_name)

    with open(theme_config, "w") as f:
        f.write(config)

    theme_nested_template_path.rename(theme_path / "templates" / new_name)
    theme_path.rename(theme_path.parent / new_name)



import logging
import os
import re
from importlib import import_module
from inspect import getmembers
from inspect import isclass
from pathlib import Path
from types import ModuleType
from typing import Dict, TextIO, Union, Optional, Any

from flask import Blueprint
from flask import Flask
from flask import session
from flask_sqlalchemy.model import DefaultMeta
from toml import load as toml_load

from .blueprint import BigAppBlueprint
from .resources import Resources
from .utilities import cast_to_bool, cast_to_import_str, deprecated


class _ModelRegistry:
    registry: Dict[str, Any]

    def __init__(self):
        self.registry = dict()

    def assert_exists(self, ref: str):
        if ref not in self.registry:
            raise KeyError(
                f"Model {ref} not found in model registry \n"
                f"Available models: {', '.join(self.registry.keys())}"
            )

    def get(self, ref: str) -> dict:
        self.assert_exists(ref)
        return self.registry[ref]

    def class_(self, ref: str) -> DefaultMeta:
        self.assert_exists(ref)
        return self.registry[ref]['class']

    def add(self, ref, model: Optional[ModuleType] = None):
        self.registry[ref] = {
            'ref': ref,
            'class': model
        }

    def __repr__(self):
        return f"ModelRegistry({self.registry})"


class BigApp(object):
    session: Dict
    themes: Dict[str, Path]

    __blueprint_registry__: Dict[str, Optional[ModuleType]]
    __model_registry__: _ModelRegistry

    _app: Flask
    _app_name: str
    _app_path: Path
    _app_folder: Path

    _default_config: Optional[str]

    def __init__(
            self,
            app: Optional[Flask] = None,
            app_config_file: Optional[str] = None
    ) -> None:

        if app is not None:
            self.init_app(app, app_config_file)

    def init_app(
            self,
            app: Flask,
            app_config_file: Optional[str] = os.environ.get("BA_CONFIG")
    ) -> None:

        """
        Initializes the application.

        -> Expects a Flask application

        Can be passed an SQLAlchemy object, doing this will enable the use of the model_class.

        Optional config from_file passed in. If no config from_file is given an
        attempt will be made to read from the environment for the variable BA_CONFIG
        If the environment variable is not found it will create a new config from_file
        called default.config.toml and proceed to load the values from there.
        """

        if app is None:
            raise ImportError(
                "No app was passed in, do ba = BigApp(flaskapp) or app.init_app(flaskapp)")
        if not isinstance(app, Flask):
            raise TypeError(
                "The app that was passed in is not an instance of Flask")

        self.session = dict()
        self.themes = dict()
        self.__model_registry__ = _ModelRegistry()

        self._app = app
        self._app_name = app.name
        self._app_path = Path(self._app.root_path)
        self._app_folder = self._app_path.parent

        if app_config_file is None:
            app_config_file = "default.config.toml"

        self._init_config(self._app_path / app_config_file)

        self.__blueprint_registry__ = dict()

    def init_session(self) -> None:
        """
        Initialize the session variables found in the config from_file.
        Use this method in the before_request route.
        """
        for key in self.session:
            if key not in session:
                session.update(self.session)
                break

    def import_builtins(self, folder: Union[str, Path] = "builtins") -> None:
        """
        Folder must be relative ( from_folder="here" not from_folder="/home/user/app/from_folder" )
        """
        if isinstance(folder, str):
            builtins_folder = Path(self._app_path / folder)
        else:
            builtins_folder = folder

        if builtins_folder.is_dir():
            with self._app.app_context():
                for py_file in builtins_folder.glob("*.py"):
                    module = import_module(
                        f"{cast_to_import_str(self._app_name, builtins_folder)}.{py_file.stem}")
                    if hasattr(module, "loader"):
                        module.loader(self._app)

    def import_blueprints(self, folder: str) -> None:
        """
        Imports all the blueprints in the given from_folder.

        Folder must be relative ( from_folder="here" not from_folder="/home/user/app/from_folder" )
        """

        folder_path = Path(self._app_path / folder)

        for potential_bp in folder_path.iterdir():
            self.import_blueprint(potential_bp)

    def import_blueprint(self, blueprint: Union[str, Path]):
        """
        Imports a single blueprint from the given path.

        Path must be relative ( path="here" not path="/home/user/app/from_folder" )
        """
        if isinstance(blueprint, str):
            potential_bp = Path(self._app_path / blueprint)
        else:
            potential_bp = blueprint

        if potential_bp.is_dir():
            if potential_bp.name in self.__blueprint_registry__:
                raise ImportError(
                    f"Blueprint {potential_bp.name} is already registered, blueprint folders must be unique\n"
                    f"Importing from {potential_bp}"
                )
            try:
                module = import_module(
                    cast_to_import_str(self._app_name, potential_bp))
                for dir_item in dir(module):
                    _ = getattr(module, dir_item)
                    if isinstance(_, BigAppBlueprint):
                        if _.enabled:
                            self.__blueprint_registry__.update(
                                {potential_bp.name: module}
                            )
                            self._app.register_blueprint(_)
                        break
            except Exception as e:
                logging.critical(f"{e}",
                                 f"Error importing blueprint: from {potential_bp}")

    @deprecated(
        "import_structures() will be removed, Use import_themes() instead")
    def import_structures(self, structures_folder: str) -> None:
        self.import_themes(structures_folder)

    def import_themes(self, themes_folder: str) -> None:
        """
        Folder must be relative ( from_folder="here" not from_folder="/home/user/app/from_folder" )
        """
        folder_path = Path(self._app_path / themes_folder)
        for theme in folder_path.iterdir():
            self.import_theme(theme)

    def import_theme(self, theme_folder: Union[str, Path]):
        """
        Folder must be relative ( from_folder="here" not from_folder="/home/user/app/from_folder" )
        """
        if isinstance(theme_folder, str):
            theme = Path(self._app_path / theme_folder)
        else:
            theme = theme_folder

        if theme.is_dir():
            if theme.name in self.themes:
                raise ImportError(
                    f"Theme {theme.name} is already registered, theme folders must be unique\n"
                    f"Importing from {theme}"
                )

            static_folder = Path(theme / "static")
            template_folder = Path(theme / "templates")
            nested_template_folder = Path(theme / "templates" / theme.name)
            builtins_folder = Path(theme / "builtins")

            config = Path(theme / "config.toml")

            if not static_folder.exists():
                logging.debug(
                    f"Static from_folder for {theme.name} was not found, skipping")

            if not template_folder.exists():
                raise NotADirectoryError(
                    f"Template from_folder for {theme.name} was not found")

            if not nested_template_folder.exists():
                raise NotADirectoryError(
                    f"Nested template from_folder for {theme.name} was not found, \nshould look like: {nested_template_folder}")

            if not config.exists():
                logging.debug(
                    f"Config from_file for {theme.name} was not found, creating default")
                config.write_text(Resources.default_theme_config.format(
                    static_url_path=f"/{theme.name}/static"))

            loaded_config = toml_load(config)

            theme_bp = Blueprint(
                name=theme.name,
                import_name=f"{theme.name}",
                static_folder=f"{theme.absolute()}/static",
                template_folder=f"{theme.absolute()}/templates",
                static_url_path=f"{loaded_config.get('static_url_path', f'/{theme.name}/static')}",
            )

            if cast_to_bool(loaded_config.get("enabled")):

                if builtins_folder.exists():
                    self.import_builtins(builtins_folder)

                self._app.register_blueprint(theme_bp)
                self.themes.update({theme.name: theme.absolute()})

    def import_models(
            self,
            from_file: Optional[Union[str, Path]] = None,
            from_folder: Optional[Union[str, Path]] = None,
    ) -> None:
        """
        Imports model files from a single from_file or a from_folder. Both are allowed to be set.

        File and Folder must be relative ( from_folder="here" not from_folder="/home/user/app/from_folder" )
        """

        def model_processor(path: Path):
            """
            Picks apart the model from_file and builds a registry of the models found.
            """
            import_string = cast_to_import_str(self._app_name, path)
            try:
                model_module = import_module(import_string)
                for model_object_members in getmembers(model_module, isclass):
                    if import_string in model_object_members[1].__module__:
                        name = model_object_members[0]
                        model = model_object_members[1]
                        if not hasattr(model, "__tablename__"):
                            raise AttributeError(
                                f"{name} is not a valid model")

                        self.__model_registry__.add(name, model)

            except ImportError as e:
                logging.critical("Error importing model: ", e,
                                 f" {import_string}")

        if from_file is None and from_folder is None:
            raise ImportError(
                "No model from_file or from_folder was passed in")

        if from_file is not None:
            if isinstance(from_file, Path):
                file_path = from_file
            else:
                file_path = Path(self._app_path / from_file)

            if file_path.is_file() and file_path.suffix == ".py":
                model_processor(file_path)
            else:
                from_folder = from_file

        if from_folder is not None:
            if isinstance(from_folder, Path):
                folder_path = from_folder
            else:
                folder_path = Path(self._app_path / from_folder)

            if folder_path.is_dir():
                for model_file in folder_path.glob("*.py"):
                    if "__" in model_file.name:
                        continue
                    model_processor(model_file)

    @deprecated("model_class() will be removed, Use model() instead")
    def model_class(self, class_name: str) -> Any:
        return self.model(class_name)

    def model(self, class_: str) -> DefaultMeta:
        """
        Returns the model class for the given class name
        """
        return self.__model_registry__.class_(class_)

    def model_meta(self, class_: Union[str, Any]) -> dict:
        """
        Returns the model class for the given class name
        """

        def check_for_table_name(model_):
            if not hasattr(model_, "__tablename__"):
                raise AttributeError(f"{model_} is not a valid model")

        if isinstance(class_, str):
            model = self.__model_registry__.get(class_)
            check_for_table_name(model['class'])
            return {
                "ref": model['ref'],
                "location": model['class'].__module__,
                "table_name": model['class'].__tablename__,
            }

        check_for_table_name(class_)
        return {
            "ref": class_.__name__,
            "location": class_.__module__,
            "table_name": class_.__tablename__,
        }

    def _init_config(self, config_file_path: Path):
        """
        Processes the values from the configuration from_file.
        """
        if not config_file_path.exists():
            logging.critical(
                f"Config from_file {config_file_path.name} was not found, creating default.config.toml to use")
            config_file_path.write_text(Resources.default_config.format(
                secret_key=os.urandom(24).hex()))

        config_suffix = ('.toml', '.tml')

        if config_file_path.suffix not in config_suffix:
            raise TypeError(
                f"Config from_file must be one of the following types: {config_suffix}")

        def if_env_replace(env_value: Optional[Any]) -> Any:
            """
            Looks for the replacement pattern to swap out values in the config from_file with environment variables.
            """
            pattern = re.compile(r'<(.*?)>')

            if isinstance(env_value, str):
                if re.match(pattern, env_value):
                    env_var = re.findall(pattern, env_value)[0]
                    return os.environ.get(env_var, "ENV_KEY_NOT_FOUND")
            return env_value

        def build_database_uri(block: dict) -> str:
            """
            Puts together the correct database URI depending on the type specified.

            Fails if type is not supported.
            """
            db_type = if_env_replace(block.get("type", "None"))
            db_name = if_env_replace(block.get('database_name', 'database'))

            db_location = if_env_replace(block.get("location", "db"))
            db_port = if_env_replace(str(block.get('port', 'None')))

            db_username = if_env_replace(block.get('username', 'None'))
            db_password = if_env_replace(block.get('password', 'None'))

            db_allowed = ('postgresql', 'mysql', 'oracle')

            if db_type == "sqlite":
                if db_location is not None:
                    if Path(db_location).exists():
                        database = Path(db_location / f"{db_name}.db")
                        return f"sqlite:///{database}"

                    db_location_path = Path(self._app_path / db_location)
                    db_location_path.mkdir(parents=True, exist_ok=True)
                    db_location_file_path = db_location_path / f"{db_name}.db"
                    return f"sqlite:///{db_location_file_path}"

                db_at_root = Path(self._app_path / f"{db_name}.db")
                return f"{db_type}:///{db_at_root}"

            if db_type in db_allowed:
                return f"{db_type}://{db_username}:{db_password}@{db_location}:{db_port}/{db_name}"

            raise ValueError(
                f"Unknown database type: {db_type}, must be: postgresql / mysql / oracle / sqlite")

        config = toml_load(config_file_path)
        flask_config = config.get("flask")
        session_config = config.get("session")
        database_config = config.get("database")

        if flask_config is not None and isinstance(flask_config, dict):
            if flask_config.get("static_folder", False):
                self._app.static_folder = if_env_replace(
                    flask_config.get("static_folder"))
                del flask_config['static_folder']

            if flask_config.get("template_folder", False):
                self._app.template_folder = if_env_replace(
                    flask_config.get("template_folder"))
                del flask_config['template_folder']

            for flask_config_key, flask_config_value in flask_config.items():
                self._app.config.update({
                    str(flask_config_key).upper(): if_env_replace(
                        flask_config_value)})

        if session_config is not None and isinstance(session_config, dict):
            self.session = session_config

        if database_config is not None and isinstance(database_config, dict):
            self._app.config['SQLALCHEMY_BINDS'] = dict()
            for database_config_key, database_config_value in database_config.items():
                if database_config_value.get("enabled", False):
                    if database_config_key == "main":
                        self._app.config[
                            'SQLALCHEMY_DATABASE_URI'] = f"{build_database_uri(database_config_value)}"
                        continue

                    self._app.config['SQLALCHEMY_BINDS'].update(
                        {
                            database_config_key: f"{build_database_uri(database_config_value)}"}
                    )

    @staticmethod
    def structure_tmpl(structure: str, template: Union[str, TextIO]) -> str:
        """
        pushes together a structure name to a template location.
        """
        return f"{structure}/{template}"



import logging
from importlib import import_module
from importlib.util import find_spec
from pathlib import Path
from types import ModuleType
from typing import Dict, Optional, Union

from flask import Blueprint
from flask import session
from toml import load as toml_load

from flask_bigapp.utilities import cast_to_bool, cast_to_import_str, deprecated


class BigAppBlueprint(Blueprint):
    """
    Class that handles Blueprints from within the Blueprint __init__ from_file
    """
    enabled: bool = False
    location: Path
    bp_name: str
    package: str
    session: dict

    nested_blueprints: Dict[str, Optional[ModuleType]]

    def __init__(self, dunder_name, config_file: str = "config.toml"):
        """
        dunder_name must be __name__
        config_file must be relative to the location of the blueprint.
        """
        self.package = dunder_name
        self.app_name = dunder_name.split(".")[0]

        spec = find_spec(self.package)

        if spec is None:
            raise ImportError(f"Cannot find origin of {self.package}")

        self.location = Path(f"{spec.origin}").parent
        self.bp_name = self.location.name

        self.nested_blueprints = {}

        bp_args = self._init_bp_config(Path(self.location / config_file)) or {}

        super().__init__(
            self.bp_name,
            self.package,
            **bp_args
        )

    def import_routes(self, folder: str = "routes") -> None:
        """
        Imports all the routes in the given from_folder.
        If no from_folder is specified defaults to a from_folder named 'routes'

        Folder must be relative ( from_folder="here" not from_folder="/home/user/app/from_folder/blueprint/from_folder" )
        """
        path = self.location / folder
        if not path.exists():
            raise NotADirectoryError(f"{path} is not a directory")

        routes = path.glob("*.py")
        for route in routes:
            try:
                import_module(f"{self.package}.{folder}.{route.stem}")
            except ImportError as e:
                logging.warning(f"Error when importing {self.package}.{route}: {e}")
                continue

    def import_nested_blueprints(self, folder: str) -> None:
        """
        Imports all the blueprints in the given from_folder.

        Folder must be relative ( from_folder="here" not from_folder="/home/user/app/from_folder" )
        """

        folder_path = Path(self.location / folder)

        for potential_bp in folder_path.iterdir():
            self.import_nested_blueprint(potential_bp)

    def import_nested_blueprint(self, blueprint: Union[str, Path]):
        """
        Imports a single blueprint from the given path.

        Path must be relative ( path="here" not path="/home/user/app/from_folder" )
        """
        if isinstance(blueprint, str):
            potential_bp = Path(self.location / blueprint)
        else:
            potential_bp = blueprint

        if potential_bp.is_dir():
            if potential_bp.name in self.nested_blueprints:
                raise ImportError(
                    f"Nested blueprint {potential_bp.name} is already registered, blueprint folders must be unique\n"
                    f"Importing from {potential_bp}"
                )
            app_name = self.package.split(".")[0]
            try:
                module = import_module(cast_to_import_str(app_name, potential_bp))
                for dir_item in dir(module):
                    _ = getattr(module, dir_item)
                    if isinstance(_, BigAppBlueprint):
                        if _.enabled:
                            self.nested_blueprints.update(
                                {potential_bp.name: module}
                            )
                            self.register_blueprint(_)
                        break
            except Exception as e:
                logging.critical(f"{e}\n", f"Error importing blueprint: from {potential_bp}")

    def init_session(self) -> None:
        """
        Initialize the session variables found in the config from_file.
        Use this method in the before_request route.
        """
        for key in self.session:
            if key not in session:
                session.update(self.session)
                break

    @deprecated("import_blueprint_models() is deprecated Use import_models instead")
    def import_blueprint_models(
            self,
            from_file: Optional[str] = None,
            from_folder: Optional[str] = None
    ) -> None:
        self.import_models(from_file=from_file, from_folder=from_folder)

    def import_models(
            self,
            from_file: Optional[Union[str, Path]] = None,
            from_folder: Optional[Union[str, Path]] = None,
    ) -> None:
        """
        Imports model files from a single from_file or a from_folder. Both are allowed to be set.

        File and Folder must be relative ( from_folder="here" not from_folder="/home/user/app/from_folder" )
        """
        from flask_bigapp import BigApp

        def get_bigapp_instance() -> BigApp:
            app_module = import_module(self.app_name)
            for dir_item in dir(app_module):
                potential_instance = getattr(app_module, dir_item)
                if isinstance(potential_instance, BigApp):
                    return potential_instance
            raise ImportError(f"Cannot find BigApp instance in {self.app_name}")

        bigapp_instance = get_bigapp_instance()

        if isinstance(from_file, str):
            from_file = Path(self.location / from_file)

        if isinstance(from_folder, str):
            from_folder = Path(self.location / from_folder)

        bigapp_instance.import_models(from_file, from_folder)

    def tmpl(self, template) -> str:
        """
        Pushes together the name of the blueprint and the template from_file to look for.
        This is a small-time saving method to allow you to only type
        bp.tmpl("index.html") when looking for template files.
        """
        return f"{self.name}/{template}"

    def _init_bp_config(self, config_file_path) -> Optional[dict]:
        """
        Attempts to load the and process the configuration from_file.
        """

        if not config_file_path.exists():
            raise FileNotFoundError(f"{self.bp_name} Blueprint config from_file {config_file_path.name} was not found")

        config_suffix = ('.toml', '.tml')

        if config_file_path.suffix not in config_suffix:
            raise TypeError(f"Config from_file must be one of the following types: {config_suffix}")

        config = toml_load(config_file_path)

        self.enabled = cast_to_bool(config.get('enabled', False))
        self.session = config.get('session', {})
        settings = config.get('settings', {})

        kwargs = dict()

        """Pull values from settings"""
        url_prefix = settings.get('url_prefix', "")
        subdomain = settings.get('subdomain', "")
        url_defaults = settings.get('url_defaults', dict())
        static_folder = settings.get('static_folder', "")
        template_folder = settings.get('template_folder', "")
        static_url_path = settings.get('static_url_path', "")
        root_path = settings.get('root_path', "")

        """If values exist in the configuration from_file, set values"""
        kwargs.update({'url_prefix': url_prefix if url_prefix != "" else f"/{self.bp_name}"})

        if subdomain:
            kwargs.update({'subdomain': subdomain})

        if url_defaults:
            kwargs.update({'url_defaults': url_defaults})

        if static_folder:
            kwargs.update({'static_folder': static_folder})

        if template_folder:
            kwargs.update({'template_folder': template_folder})

        if static_url_path:
            kwargs.update({'static_url_path': static_url_path})

        if root_path:
            kwargs.update({'root_path': root_path})

        return kwargs



from .auth import Auth as Auth
from .auth import PasswordGeneration as PasswordGeneration
from .bigapp import BigApp as BigApp
from .blueprint import BigAppBlueprint as Blueprint
from .security import BigAppSecurity as Security

__all__ = ["Auth", "PasswordGeneration", "BigApp", "Blueprint", "Security"]



import pathlib
from setuptools import setup, find_packages

readme = pathlib.Path(pathlib.Path.cwd() / "README.md").read_text()

setup(
    name='Flask-BigApp',
    version=f'2023.1.6',
    url='https://github.com/CheeseCake87/Flask-BigApp',
    license='GNU Lesser General Public License v2.1',
    author='David Carmichael',
    author_email='carmichaelits@gmail.com',
    description='A Flask auto importer that allows your Flask apps to grow big.',
    long_description=f'{readme}',
    long_description_content_type='text/markdown',
    platforms='any',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment', 'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)',
        'Operating System :: OS Independent', 'Programming Language :: Python',
        'Programming Language :: Python :: 3.8',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Natural Language :: English',
    ],
    install_requires=[
        'Flask',
        'Flask-SQLAlchemy',
        'toml',
    ],
    zip_safe=False,
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'flask-bigapp = flask_bigapp_cli:cli',
        ]
    }
)



def test_general_setup(client):
    """
    If this test is successful, the app is set up and running correctly.
    """
    response = client.get('/tests/')
    assert response.status_code == 200


def test_error_404(client):
    """
    If this test is successful, the app is set up and running correctly.
    """
    response = client.get('/tests/error-page-404')
    assert b"No route associated with the URL" in response.data


def test_error_500(client):
    """
    If this test is successful, the app is set up and running correctly.
    """
    response = client.get('/tests/error-page-500')
    assert b"There has been a server error!" in response.data


def test_context_processors(client):
    """
    If this test is successful, the context processors are working correctly.
    """
    response = client.get('/tests/context-processors')
    assert b"$100.00" in response.data


def test_import_theme(client):
    """
    If this test is successful, the import of the theme is working correctly.
    """
    response = client.get('/tests/static')
    assert b"/img/Flask-BigApp-Logo.png" in response.data


def test_database_creation(client):
    """
    If this test is successful, the database has been dropped and created correctly.
    """
    response = client.get('/tests/database-creation')
    assert b"Database created." in response.data


def test_database_population(client):
    """
    If this test is successful, the database has been populated correctly.
    """
    response = client.get('/tests/database-population')
    assert b"David created" in response.data
